/**
 * Process structure extraction & definition.
 */

import fs from 'fs'
import path from 'path'
import async from 'async'
import mkdirp from 'mkdirp'
import computeAppId from '../compute-app-id'
import createArrayStore from './array-store'
import processThresholds from './thresholds'
import createThresholdActionQueue from './threshold/action-queue'

import { ModelGraph } from './models'

const debug = require('debug')('nsolid-console/cache')

const utf8enc = { encoding: 'utf8' }

export default class Cache {
  constructor (config, previousCache) {
    if (typeof config === 'string') {
      // In test, storage path is passed as first arg.
      // rather than rewrite all tests, can simply handle this case.
      config = {
        storage: config
      }
    }

    this.getApp = createArrayStore({
      hosts: [],
      processes: []
    })

    // ensure rc.securityConfig object exists
    const rc = config.rc || {}
    const securityConfig = rc.securityConfig || {auto: true, enabled: true, recipients: []}

    this.config = config
    this.previousCache = previousCache || null
    this.processToHub = {}
    this.hosts = {}
    this.fileResources = {}
    this.storagePath = config.storage
    this.appFile = path.join(this.storagePath, 'applications.json')
    this.graph = (previousCache && previousCache.graph) || new ModelGraph()
    this.security = (previousCache && previousCache.security) || {
      updating: false,
      updated: 0, // datestamp. 0 == never
      sentNotifications: {}
    }
    this.fileResourceWatcher = null

    // values must always come from config
    Object.assign(this.security, { recipients: [] }, {
      recipients: securityConfig.recipients,
      enabled: securityConfig.enabled,
      auto: securityConfig.auto
    })

    if (!previousCache) {
      this.thresholdActionQueue = createThresholdActionQueue(config)
    }
  }

  /**
   * Add a threshold action to the threshold action queue
   */
  addThresholdAction (action) {
    if (this.previousCache) {
      return this.previousCache.addThresholdAction(action)
    }

    this.thresholdActionQueue.push(action)
  }

  /**
   * Serialize and store cache contents on disk
   */
  save (done) {
    fs.writeFile(this.appFile, JSON.stringify(this.toJSON(true)), utf8enc, done)
  }

  /**
   *  Restore the cache contents from disk
   */
  restore (done) {
    if (!done) {
      throw new Error('Cache#restore requires a passed callback')
    }

    fs.readFile(this.appFile, utf8enc, (e, json) => {
      if (e) {
        return done(e)
      }

      try {
        const obj = JSON.parse(json)
        this.fromJSON(obj, done)
      } catch (e) {
        return done(e)
      }
    })
  }

  /**
   * Each polling cycle gets its own cache so we can perform diffing and similar
   */
  collectPollingResult (fn, done) {
    if (!done) {
      throw new Error('Cache#collectPollingResult requires a passed callback')
    }
    const cache = new Cache(this.config, this)

    // each polling cycle
    fn(cache)

    // merge the polling cache
    this.processToHub = cache.processToHub
    this.getApp = cache.getApp
    this.hosts = cache.hosts

    this.filterFileResourcesByDiskPresence(this.fileResources, (filterError, filteredFileResources) => {
      if (filterError) {
        return done(filterError)
      }

      this.fileResources = filteredFileResources

      // ensure app order is always the same
      this.getApp.array.sort(function (a, b) {
        return a.name.localeCompare(b.name)
      })

      // TODO: compute diff from previous cycle and pass it along

      done()
    })
  }

  /**
   * Extract process structure from process & host.
   * Mutates Application instance gathered from getApp.
   *
   * @param {Process} proc
   * @param {Host} host
   * @param {Function} getApp pass app name to retrieve Application instance
   * @return {Application}
   * @return {rebuildActionQueue} set to true to rebuild the threshold action queue on boot
   */
  addProcess (proc, time, rebuildActionQueue) {
    const application = this.getApp(proc.app, {
      id: computeAppId(proc.app),
      name: proc.app,
      time: time || Date.now()
    })

    // cache what proxy the process was collected on
    this.processToHub[proc.id] = proc.hub

    // process thresholds
    const config = this.config
    if (config && config.rc && config.rc.thresholds) {
      const thresholds = config.rc.thresholds
      const appid = computeAppId(proc.app)
      if (thresholds && thresholds.app && thresholds.app[appid]) {
        processThresholds(proc, this, thresholds.app[appid], rebuildActionQueue)
      }
    }

    application.processes.push(proc)

    // only insert one instance of the host object per application.
    if (application.hosts.indexOf(proc.host) === -1) {
      application.hosts.push(proc.host)
    }

    return application
  }

  /**
   * Get a process by app and process id
   *
   * @param {String} appid
   * @param {String} processid
   * @return {Process}
   */
  getProcess (appid, processid) {
    const app = this.getApp(appid)
    if (!app) {
      return null
    }

    const processes = app.processes
    const l = processes.length
    for (let i = 0; i < l; i++) {
      const proc = app.processes[i]
      if (proc.id === processid) {
        return proc
      }
    }

    return null
  }

  /**
   * Add metadata to cache describing a resource (snapshot/profile)
   *
   * @param {String} type
   * @param {Object} meta
   * @return undefined
   */
  addFileResource (type, meta) {
    // always add file resources to the top level cache
    if (this.previousCache) {
      return this.previousCache.addFileResource(type, meta)
    }

    if (!this.fileResources[type]) {
      this.fileResources[type] = {}
    }

    const resource = this.fileResources[type]
    const app = computeAppId(meta.app)
    if (!resource[app]) {
      resource[app] = []
    }

    // TODO: track this as an addition
    resource[app].unshift(meta)
  }

  generateResourceFilename (appName, processId, time, extension) {
    var appId = computeAppId(appName)
    if (!appId) {
      return null
    }
    return `${appId}-${processId}-${time}.${extension}`
  }

  /**
   * Save a resource to disk while adding a metadata entry to the cache
   *
   * @param {String} type
   * @param {Object} meta
   * @return {WritableStream}
   */
  createFileResourceWriteStream (type, meta, done) {
    if (!done) {
      throw new Error('Cache#createFileResourceWriteStream requires a passed callback')
    }

    const base = path.join(this.storagePath, type)
    const dest = path.join(base, meta.file)
    const cache = this

    mkdirp(base, (err) => {
      if (err) {
        return done(err)
      }

      const stream = fs.createWriteStream(dest)

      stream.on('finish', () => {
        cache.addFileResource(type, meta)
      })

      done(null, stream)
    })
  }

  /**
   * Create a resource metadata object to be used with
   *   - createFileResourceWriteStream
   *   - addFileResource
   *
   * @param {Process} proc
   * @param {String}  fileExtension
   * @param {Integer} now (optional)
   * @return {Object}
   */
  getResourceMetaFromProcess (proc, fileExtension, now) {
    now = now || Date.now()
    const appid = computeAppId(proc.app)
    const processid = proc.id
    return {
      time: now,
      app: proc.app,
      id: proc.id,
      file: this.generateResourceFilename(appid, processid, now, fileExtension),
      title: proc.title,
      pid: proc.pid
    }
  }

  getHubByProcessId (processId) {
    const hub = this.processToHub[processId]
    if (!hub) {
      return null
    }

    return hub
  }

  appIdFromFileResourceFilename (filename) {
    // dog-food-direct-f76123df123-123123123.heapsnapshot -> dog-food-direct
    return String(filename).split('-').slice(0, -2).join('-')
  }

  /**
   * Mutates this.fileResources[type] and removes any entries that do not
   * have counterparts on the filesystem
   */
  filterFileResourcesByDiskPresence (obj, done) {
    const types = ['profiles', 'snapshots']
    obj = Object.assign({ profiles: {}, snapshots: {} }, obj)

    const ret = { profiles: {}, snapshots: {} }

    async.eachSeries(types, (type, next) => {
      if (!ret[type]) {
        debug('weird type passed to Cache#filterFileResourcesByDiskPresence: %s', type)
        ret[type] = {}
        return next()
      }

      const resourceDir = path.join(this.storagePath, type)
      fs.readdir(resourceDir, (readDirError, files) => {
        // ignore this error as the idea is to simply reset the list of
        // file resources under this type to an empty set
        if (readDirError) {
          debug('file resource directory does not exist, marking metadata as empty (%s)', type)
          ret[type] = {}
          return mkdirp(resourceDir, (e) => {
            if (e) {
              debug('unable to create resource directory: %s - %s', resourceDir, e.message)
            }

            next()
          })
        }

        // compute a map between the appId and filenames that live on disk
        const fileResourceByApplication = {}

        files.forEach((file) => {
          const appId = this.appIdFromFileResourceFilename(file)

          if (!fileResourceByApplication[appId]) {
            fileResourceByApplication[appId] = new Set()
          }

          fileResourceByApplication[appId].add(file)
        })

        // filter away all file resource metadata that does not exist on disk
        const applications = obj[type]

        Object.keys(applications).forEach((appId) => {
          if (!fileResourceByApplication[appId]) {
            debug('application does not exist in cache, ignoring associated metadata')
            return
          }

          ret[type] = obj[type]

          const resourcesByApp = fileResourceByApplication[appId]
          applications[appId] = applications[appId].filter((metadata) => {
            if (!metadata || !metadata.file) {
              return false
            }
            return resourcesByApp.has(metadata.file)
          })
        })

        next()
      })
    }, (e) => {
      done(e, ret)
    })
  }

  /**
   * Restore the cache from disk
   */
  fromJSON (obj, done) {
    if (!done) {
      throw new Error('Cache#fromJSON requires a passed callback')
    }

    if (!obj) {
      return done()
    }

    this.filterFileResourcesByDiskPresence(obj, (filterError, filteredFileResources) => {
      if (obj.hosts) {
        this.hosts = obj.hosts
      }

      this.fileResources = filteredFileResources

      if (obj.applications && obj.applications.length) {
        obj.applications.forEach((app) => {
          app.processes.forEach((proc) => {
            // add the process and signal that we want the action queue to be rebuilt
            this.addProcess(proc, app.time, true)
          })
        })
      }

      if (obj.graph) {
        ModelGraph.import(this.graph, new ModelGraph(obj.graph))
      }

      if (obj.security) {
        const rc = this.config.rc || {}
        const securityConfig = rc.securityConfig || {auto: true, enabled: true, recipients: []}
        Object.assign(this.security, obj.security, {
          recipients: securityConfig.recipients,
          enabled: securityConfig.enabled,
          auto: securityConfig.auto,
          updating: false // never restore updating state
        })
      }

      done()
    })
  }

  toJSON () {
    return Object.assign({
      hosts: this.hosts,
      applications: this.getApp.array,
      snapshots: {},
      profiles: {},
      graph: this.graph.export(),
      security: this.security
    }, this.fileResources)
  }
}
