'use strict'

import { Graph, json } from 'graphlib'
import { EventEmitter } from 'events'

const TYPE_EDGE = 'type'

class DataGraph {
  constructor ({graph, data} = {}) {
    // import data
    this.data = data || {}
    this.graph = this.initGraph(graph)

    this.types = {}
    this.relations = new Graph({directed: true})

    this.pruning = new Map()
    EventEmitter.call(this)
  }

  initGraph (graph) {
    return (
      !graph
      ? new Graph({directed: true, multigraph: true})
      : json.read(graph)
    )
  }

  export () {
    return {
      data: this.data,
      graph: json.write(this.graph)
    }
  }

  addType (name, Type) {
    if (typeof name !== 'string') throw new TypeError('DataGraph#addType: expecting String name')
    if (!Type.id) throw new TypeError('DataGraph#configureType: Type constructor must have id function')
    this.types[name] = Type
    this.graph.setNode(name)
  }

  getTypes () {
    return Object.keys(this.types)
  }

  getType (type) {
    const Type = this.types[type]
    if (!Type) throw new ReferenceError(`DataGraph: Unknown type: ${type}`)
    return Type
  }

  add (type, nodeDatas, previousDatas) {
    previousDatas = previousDatas || {}

    const Type = this.getType(type)
    if (!Array.isArray(nodeDatas)) throw new TypeError(`DataGraph#add: expecting an Array of Nodes, got ${typeof nodeDatas}`)

    // get only unique nodes, keep last occurance in array
    const nodes = getUniqueNodes(Type, nodeDatas)
    .map(data => new Type(data))

    // check all ids are strings
    checkIDs(nodes.map(n => n.id))

    // find new nodes
    const newNodes = nodes
    .filter(node => this.data[node.id] === undefined)

    // setup type
    newNodes.forEach(node => {
      this.graph.setEdge(type, node.id, TYPE_EDGE)
    })

    // update data for all nodes, not just new nodes
    nodes.forEach(node => {
      this.data[node.id] = Object.assign(previousDatas[node.id] || {}, node)
    })

    if (newNodes.length) {
      // allow relations etc to be set up syncronously
      // before triggering events
      setImmediate(() => this.emit(`added:${type}`, newNodes))
    }

    return newNodes
  }

  /**
   * Replace existing set of nodes.
   *
   * Adds new nodes found in `nodes`
   * Removes existing nodes not found in `nodes`
   */

  replace (type, nodes) {
    const Type = this.getType(type)
    if (!Array.isArray(nodes)) throw new TypeError(`DataGraph#replace: expecting an Array of Nodes, got ${typeof nodes}`)

    const activeNodeIds = new Set(nodes.map(node => Type.id(node)))

    // find current nodes that aren't in active node list
    const idsToRemove = this.getIds(type)
    .filter(id => !activeNodeIds.has(id))

    const previousDatas = {}
    activeNodeIds.forEach((id) => {
      previousDatas[id] = this.data[id]
    })

    this.remove(type, idsToRemove)
    this.add(type, nodes, previousDatas)
  }

  /**
   * Get all nodes for a particular `type`.
   */

  getNodes (type) {
    return this.getIds(type)
    .map(id => this.data[id])
  }

  /**
   * Get all ids for a particular `type`.
   */

  getIds (type) {
    this.getType(type)
    return this.graph.successors(type)
  }

  /**
   * Remove all nodes of `type` matching `ids`.
   * Fires 'removed:${type}' event with Array of removed node data.
   * Will not fire event if nothing was removed.
   */

  remove (type, ids) {
    this.getType(type) && checkIDs(ids)
    const pruningType = this.pruning.get(type)
    if (pruningType) ids = ids.filter(id => !pruningType.has(id))

    const nodesToRemove = this.getData(ids)
    if (!nodesToRemove.length) return []
    this.emit(`before:removed:${type}`, nodesToRemove)
    nodesToRemove.forEach(node => {
      this.graph.removeNode(node.id)
      delete this.data[node.id]
      return node
    })

    if (nodesToRemove.length) {
      setImmediate(() => this.emit(`removed:${type}`, nodesToRemove))
    }

    return nodesToRemove
  }

  prune (idsBeingRemoved, fromType, toType) {
    const startedPruning = !this.pruning.size
    const type = this.pruning.get(fromType) || new Set()
    this.pruning.set(fromType, new Set([...type, ...idsBeingRemoved]))
    idsBeingRemoved.forEach(id => {
      // remove related items that only have references that are being removed
      const relatedToTypeIds = this.related(id, fromType, toType)
      const itemsToPrune = relatedToTypeIds.filter(relatedId => {
        const relatedFromTypeIds = this.relatedBack(relatedId, toType, fromType)
        // if has other items pointing at it that aren't about to be removed
        return relatedFromTypeIds.every(relatedFromId => idsBeingRemoved.indexOf(relatedFromId) !== -1)
      })
      this.remove(toType, itemsToPrune)
    })

    if (!startedPruning) return
    this.pruning = new Map()
  }

  addWeakRelation (fromType, toType) {
    this.addRelation(fromType, toType)
    this.pruneOnRemoved(fromType, toType)
  }

  pruneOnRemoved (fromType, toType) {
    this.on(`before:removed:${fromType}`, itemsBeingRemoved => {
      this.prune(itemsBeingRemoved.map(item => item.id), fromType, toType)
    })
  }

  /**
   * Get data associated with each id in `ids`.
   */

  getData (ids) {
    checkIDs(ids)
    return ids.map(id => this.data[id]).filter(Boolean)
  }

  /**
   * Remove relation between ids `from` and `to`.
   */

  unrelate (from, to) {
    checkIDs([from, to])
    const edge = this.graph.edge(from, to)
    if (!edge) return false
    this.graph.removeEdge(from, to)
    return true
  }

  /**
   * Create relation between ids `from` and `to`,
   * where ids of of types `fromType` & `toType` respectively.
   */

  relate (fromType, toType, from, to) {
    this.getType(fromType) && this.getType(toType) && checkIDs([from, to])
    const relation = this.getRelation(fromType, toType)
    if (!relation) throw new TypeError(`DataGraph#relate: There is no relation from ${fromType} to ${toType}`)

    // do nothing if ids do not exist
    if (this.data[from] === undefined || this.data[to] === undefined) return

    // don't relate twice
    const edge = this.graph.edge(from, to)
    if (edge && edge === relation) return
    return this.graph.setEdge(from, to, relation)
  }

  /**
   * Create a relationship between two types.
   */

  addRelation (fromType, toType) {
    this.getType(fromType) && this.getType(toType)

    const label = `${fromType} has ${toType}`
    const edge = this.relations.edge(fromType, toType)
    if (edge && edge === label) return

    this.relations.setEdge(fromType, toType, label)
    return true
  }

  /**
   * Get the object representing the relationship between
   * the types `fromType` & `toType`.
   */

  getRelation (fromType, toType) {
    this.getType(fromType) && this.getType(toType)
    return this.relations.edge(fromType, toType)
  }

  /**
   * Get types with a relationship to `fromType`.
   */

  getRelatedTypes (fromType) {
    this.getType(fromType)
    const types = (this.relations.outEdges(fromType) || [])
    .map(({w}) => w)
    return Array.from(new Set(types))
  }

  /**
   * Get ids related to `fromId` across a `fromType` to `toType` relationship.
   *
   * e.g. related(appid, 'applications', 'processes') would give all processes
   * related to the given application.
   */

  related (fromId, fromType, toType) {
    return this._related(fromId, fromType, toType)
  }

  /**
   * Same as related, but for cases where the relationship is reversed.
   */

  relatedBack (fromId, fromType, toType) {
    return this._related(fromId, fromType, toType, true)
  }

  _related (fromId, fromType, toType, reversed = false) {
    this.getType(fromType) && this.getType(toType) && checkID(fromId)

    // if fromId doesn't exist can't do anything
    if (!this.data[fromId]) return []

    const relation = reversed ? this.getRelation(toType, fromType) : this.getRelation(fromType, toType)

    if (!relation) return []

    const edges = reversed
    ? this.graph.inEdges(fromId)
    : this.graph.outEdges(fromId)

    const related = edges
    .filter(edge => this.graph.edge(edge.v, edge.w) === relation)
    .map(edge => reversed ? edge.v : edge.w)

    return Array.from(new Set(related))
  }
}

/**
 * Import data and relationhips from `sourceGraph` into `targetGraph`.
 * Makes correct added/removed events fire on targetGraph
 */

DataGraph.import = function (targetGraph, sourceGraph) {
  // add all nodes first
  var sourceTypes = sourceGraph.getTypes()
  var sourceNodes = {}
  var targetNodes = {}

  var i = 0
  var j = 0
  var type = null

  // get nodes from every type in sourceGraph
  for (i = 0; i < sourceTypes.length; i++) {
    type = sourceTypes[i]
    sourceNodes[type] = sourceGraph.getNodes(type)
    targetNodes[type] = targetGraph.getNodes(type)
    targetGraph.add(type, sourceNodes[type])
  }

  // remove nodes that no longer exist
  for (i = 0; i < sourceTypes.length; i++) {
    type = sourceTypes[i]
    var removeNodes = []
    for (j = 0; j < targetNodes[type].length; j++) {
      var targetNode = targetNodes[type][j]
      if (sourceNodes[type].find((sourceNode) => sourceNode.id === targetNode.id)) continue
      removeNodes.push(targetNode.id)
    }

    targetGraph.remove(type, removeNodes)
  }

  for (i = 0; i < sourceTypes.length; i++) {
    type = sourceTypes[i]
    targetGraph.add(type, sourceNodes[type])
  }

  // setup relations after all nodes added
  for (i = 0; i < sourceTypes.length; i++) {
    type = sourceTypes[i]
    var relatedTypes = sourceGraph.getRelatedTypes(type)
    for (j = 0; j < relatedTypes.length; j++) {
      var relatedType = relatedTypes[j]
      for (var k = 0; k < sourceNodes[type].length; k++) {
        var node = sourceNodes[type][k]
        var relatedIds = sourceGraph.related(node.id, type, relatedType)
        for (var m = 0; m < relatedIds.length; m++) {
          var relatedId = relatedIds[m]
          targetGraph.relate(
            type, relatedType,
            node.id, relatedId
          )
        }
      }
    }
  }
}

function checkIDs (ids) {
  if (!Array.isArray(ids)) throw new TypeError(`DataGraph: expecting an Array of String ids, got ${typeof ids}`)
  for (let i = 0; i < ids.length; i++) {
    checkID(ids[i])
  }
  return ids
}

function checkID (id) {
  if (typeof id !== 'string') throw new TypeError(`DataGraph: Expected String id, got ${typeof id}`)
  return id
}

/**
 * Get only unique nodes, keeps last occurance in array
 */

function getUniqueNodes (Type, nodes) {
  const uniqueIds = new Set()
  return nodes.slice().reverse() // iterate backwards to keep last occurance of duped node
  .filter(node => {
    const id = Type.id(node)
    if (uniqueIds.has(id)) return false
    uniqueIds.add(id)
    return true
  })
  .reverse() // restore order
}

Object.assign(DataGraph.prototype, EventEmitter.prototype)

module.exports = DataGraph
