/**
 * Wrapper interface for executing N|Solid CLI commands
 */

import split from 'split'
import async from 'async'
import through from 'through'
import Client from 'nsolid-rpcclient'
import json from './json'
import getConfig from './config'
import concat from 'concat-stream'
import combineErrors from 'combine-errors'

module.exports.broadcast = broadcast
module.exports.command = command
module.exports.createStream = createStream

function createStream (cmd, args, hub, config) {
  config = config || getConfig.defaults
  const timeout = (args && args.timeout) || config.timeout

  const passthrough = through()

  try {
    const client = new Client(hub)
    let s = null

    if (args && args.id) {
      s = client.routeTo(cmd, args, null, timeout)
    } else {
      s = client.send(cmd, null, timeout)
    }

    // if the stream is not piped, provide a noop destination
    // so things flow instead of buffering and never ending
    let piped = false
    s.once('pipe', () => piped = true)
    process.nextTick(() => {
      if (!piped) {
        s.pipe(passthrough)
      }
    })
    return s
  } catch (e) {
    process.nextTick(() => {
      passthrough.emit('error', e)
    })
    return passthrough
  }
}

function command (cmd, args, hub, config, fn) {
  const stream = createStream(cmd, args, hub, config)

  stream.pipe(concat((buf) => {
    const obj = json(buf.toString())
    if (obj.error) {
      return fn(new Error(obj.error))
    }

    fn(null, obj)
  }))

  stream.on('error', fn)
}

function broadcast (cmd, config, fn) {
  function concatAndRespond (e, res) {
    if (e) return fn(e)

    const valid = []
    const errors = []

    res.forEach((r) => {
      if (r.error) {
        errors.push(r)
      } else {
        Array.prototype.push.apply(valid, r)
      }
    })

    fn(errors.length ? errors : null, valid)
  }

  function broadcastToHub (hub, done) {
    const s = createStream(cmd, null, hub, config)
    const res = []

    s.on('error', (e) => {
      done(null, {
        error: e,
        hub: hub
      })
    })

    s.pipe(split()).on('data', function gotLine (line) {
      var obj = json((line || '').toString())
      if (!obj.error) {
        if (obj.meta) {
          obj.meta.hub = hub
        }
        res.push(obj)
      }
    })

    s.on('end', function streamEnded () {
      done(null, res)
    })
  }

  async.map(config.hub, broadcastToHub, concatAndRespond)
}

module.exports.commandMany = commandMany

function commandMany (cmd, processes, config = getConfig.defaults, done) {
  async.map(processes, (process, next) => {
    const hub = process.hub
    // processes require a hub
    if (!hub) return next(new Error('commandMany: process must have a hub'))
    const client = new Client(hub)
    client.routeToBuffered(cmd, process, null, (err, res, body) => {
      if (err) return next(null, err) // do not stop on error
      if (res.statusCode >= 300) {
        return next(null, new Error(`Request Error: ${res.status} ${res.url}:\n${body}`))
      }
      try {
        var data = LDJSONParse(String(body))
      } catch (err) {
        return next(null, err)
      }
      if (data.length > 1) return next(null, new Error('commandMany assumes single response item'))
      return next(null, {
        meta: process,
        reply: data[0] // assumes only ever a single response entry
      })
    }, config.timeout)
  }, (err, results) => {
    if (err) return done(err)
    const errs = results.filter(r => r instanceof Error)
    if (!errs.length) return done(null, results)
    results = results.filter(r => !(r instanceof Error))
    done(combineErrors(errs.filter(e => !e.message.includes('Unable to locate host'))), results)
  })
}

function LDJSONParse (str = '{}') {
  return JSON.parse(('[' + str.replace(/}\n/gm, '},') + ']').replace(/,]$/, ']'))
}
