import DataGraph from './data-graph'
import computeAppId from '../compute-app-id'
import sha from 'stable-sha1'

const SEP = '|' // non-conflicting separator between package name & version

/*
 * Each model class below exposes a static id function which produce a
 * deterministic identifier based on raw data passed to them from the hub. This
 * is a requirement for Constructors passed to DataGraph. DataGraph uses this to
 * determine if a piece of incoming data is unique or not, without the need to
 * instantiate a new instance.
 */

export class Package {

  /**
   * Generate an ID from raw package data.
   */

  static id (pkg) {
    if (!pkg.name) throw new Error('Package has no name!')
    if (!pkg.version) throw new Error('Package has no version!')

    return pkg.name + SEP + pkg.version
  }

  constructor (pkg) {
    this.id = Package.id(pkg)
    this.name = pkg.name
    this.version = pkg.version
    this.modules = null // to be set to canonical list from server

    // would be nice to have this, but packages w/same name/version
    // seem to be overwritten in the graph, so this basically gets
    // thrown away for cases where name/version match, but path
    // doesn't ...
    //
    // this.path = pkg.path
  }
}

export class Process {

  /**
   * Generate an ID from raw Process data.
   */

  static id (process) {
    if (!process.id) throw new Error('Process has no id!')

    return process.id
  }

  constructor (process) {
    Object.assign(this, process)
    this.id = Process.id(process)
  }

}

export class Vulnerability {

  /**
   * Generate an ID from raw vulnerability data.
   */

  static id (vuln) {
    if (!vuln.id) throw new Error('Vulnerability has no ids!')

    return vuln.id
  }

  constructor (vulnerability) {
    Object.assign(this, vulnerability)
    this.id = Vulnerability.id(vulnerability)
  }
}

export class Application {

  /**
   * Generate an ID from raw application data.
   */

  static id (application) {
    return computeAppId(application.name)
  }

  constructor ({name}) {
    this.name = name
    this.id = Application.id({name})
  }
}

/**
 * Canonical list of modules for a package.
 */

export class ModuleList {

  /**
   * Generate an ID from a package module list.
   */

  static id (moduleList) {
    return Package.id(moduleList) + SEP + 'ModuleList'
  }

  constructor (moduleList) {
    this.name = moduleList.name
    this.version = moduleList.version
    this.id = ModuleList.id(moduleList)
    const modules = moduleList.modules || []
    this.modules = modules.map(({path, shasum}) => ({path, shasum}))
  }
}

/**
 * Container for modules which have been modified in a package.
 * TODO: Merge ModuleList, ModifiedPackage & perhaps Package into single class.
 * Multiple classes required at the moment because datagraph needs to add support
 * for multiple different types of relationships between the same nodes.
 */

export class ModifiedPackage {

  /**
   * Generate an ID from modified package info.
   */

  static id (modifiedPkg) {
    if (modifiedPkg.id) return modifiedPkg.id
    return modifiedPkg.packageId + sha(modifiedPkg)
  }

  constructor (modifiedPkg) {
    this.packageId = modifiedPkg.packageId
    this.path = modifiedPkg.path
    this.modules = modifiedPkg.modules
    this.id = ModifiedPackage.id(modifiedPkg)
  }
}

/**
 * Initialises specific types relationships for application
 * and exposes domain-specific utility functions.
 */

export class ModelGraph extends DataGraph {
  constructor ({graph, data} = {}) {
    super({graph, data})
    this.addType('applications', Application)
    this.addType('processes', Process)
    this.addType('packages', Package)
    this.addType('vulnerabilities', Vulnerability)
    this.addType('modulelists', ModuleList)
    this.addType('modifiedpackages', ModifiedPackage)
    this.addWeakRelation('processes', 'applications')
    this.addWeakRelation('processes', 'modifiedpackages')
    this.addWeakRelation('processes', 'packages')
    this.addWeakRelation('packages', 'vulnerabilities')
    this.addWeakRelation('packages', 'modulelists')
  }

  /**
   * Add relationship between process & list of modified packages.
   */

  addModifiedPackages (processId, modifiedPackages = []) {
    if (!modifiedPackages.length) return
    modifiedPackages = modifiedPackages.map(modifiedPackage => {
      // premptively calculate id, avoids double handling
      return Object.assign(modifiedPackage, {id: ModifiedPackage.id(modifiedPackage)})
    })
    this.add('modifiedpackages', modifiedPackages)
    return modifiedPackages.filter(modifiedPackage => {
      return this.relate('processes', 'modifiedpackages', processId, modifiedPackage.id)
    })
  }

  getApplicationSecurityUpdateTime (app) {
    const id = Application.id(app)
    if (!this.data[id]) return

    return this.data[id].lastSecurityUpdate
  }

  /**
   * Get Package IDs of modified packages for a process.
   */

  modifiedPackagesIds (processId) {
    return (
      this.getData(this.related(processId, 'processes', 'modifiedpackages'))
      .map(m => m.packageId)
    )
  }

  /**
   * Get Package IDs of vulnerable AND modified packages for a process.
   */

  modifiedVulnerablePackages (processId) {
    const modifiedPackages = this.modifiedPackagesIds(processId)
    return modifiedPackages.filter(({packageId}) => this.packageVulnerabilities(packageId).length)
  }

  /**
   * Get canonical package information for packageId.
   */

  getCanonicalPackage (packageId) {
    return this.data[this.related(packageId, 'packages', 'modulelists')]
  }

  /**
   * Get all processes for an application
   */

  applicationProcesses (appId) {
    return this.relatedBack(appId, 'applications', 'processes')
  }

  /**
   * Get all packages for a process
   */

  processPackages (processId) {
    return this.related(processId, 'processes', 'packages')
  }

  /**
   * Get vulnerabilities for an application
   */

  applicationVulnerabilities (id) {
    const results = this.relatedBack(id, 'applications', 'processes')
    .map(id => this.processVulnerabilities(id))
    .reduce(flatten, [])
    return Array.from(new Set(results))
  }

  activeVulnerabilities () {
    return unique(
      this.getIds('applications')
      .map(appId => this.applicationVulnerabilities(appId))
      .reduce(flatten, [])
      .filter(vulnId => this.vulnerabilityProcesses(vulnId).length) // has at least one affected process
    )
  }

  /**
   * Get vulnerabilities for an application
   */

  processVulnerabilities (id) {
    const results = this.related(id, 'processes', 'packages')
    .map(id => this.related(id, 'packages', 'vulnerabilities'))
    .reduce(flatten, [])
    return Array.from(new Set(results))
  }

  vulnerabilityProcesses (id) {
    const results = this.relatedBack(id, 'vulnerabilities', 'packages')
    .map(id => this.relatedBack(id, 'packages', 'processes'))
    .reduce(flatten, [])
    return Array.from(new Set(results))
  }

  vulnerabilityApplications (id) {
    const results = this.vulnerabilityProcesses(id)
    .map(id => this.related(id, 'processes', 'applications'))
    .reduce(flatten, [])
    return Array.from(new Set(results))
  }

  /**
   * Get package id for a vulnerability.
   * Assumes one package per vulnerability.
   */

  vulnerabilityPackage (vulnId) {
    return this.relatedBack(vulnId, 'vulnerabilities', 'packages')[0]
  }

  /**
   * Get package data for a vulnerability.
   * Assumes one package per vulnerability.
   */

  vulnerabilityPackageData (vulnId) {
    return this.data[this.vulnerabilityPackage(vulnId)]
  }

  /**
   * Get all packages for a vulnerability.
   */

  vulnerabilityPackages (vulnId) {
    return this.relatedBack(vulnId, 'vulnerabilities', 'packages')
  }

  /**
   * Get all packages for a vulnerability in a process.
   */

  vulnerabilityProcessPackages (vulnId, processId) {
    const processPackages = new Set(this.processPackages(processId))
    return this.vulnerabilityPackages(vulnId)
    .filter(pkgId => processPackages.has(pkgId))
  }

  /**
   * Get all packages with a vulnerability
   */

  vulnerablePackages (ids) {
    const vulnerabilities = ids || this.getIds('vulnerabilities')
    return unique(
      vulnerabilities.map(id => this.relatedBack(id, 'vulnerabilities', 'packages'))
      .reduce(flatten, [])
    )
  }

  /**
   * Get all processes with a vulnerability
   */

  vulnerableProcesses () {
    return unique(
      this.vulnerablePackages()
      .map(id => this.relatedBack(id, 'packages', 'processes'))
      .reduce(flatten, [])
    )
  }

  /**
   * Get all processes with vulnerabilities for an application
   */

  vulnerableProcessesForApplication (appId) {
    const vulnerableProcesses = new Set(this.vulnerableProcesses())
    return this.applicationProcesses(appId)
    .filter(processId => vulnerableProcesses.has(processId))
  }

  /**
   * Get all processes in an application with a particular vulnerability.
   */

  vulnerabilityApplicationProcesses (vulnId, appId) {
    const vulnerabilityProcesses = new Set(this.vulnerabilityProcesses(vulnId))
    return this.applicationProcesses(appId)
    .filter(processId => vulnerabilityProcesses.has(processId))
  }

  /**
   * Get all processes using packages.
   */

  packagesProcesses (packageIds) {
    return unique(
      packageIds.map(pkgId => this.relatedBack('processes', 'packages', pkgId))
      .reduce(flatten, [])
    )
  }

  /**
   * Get all vulnerabilities for a package
   */

  packageVulnerabilities (pkgId) {
    return this.related('packages', 'vulnerabilities', pkgId)
  }
}

/**
 * Array#reduce util. Flattens a singly nested Array.
 */

function flatten (all, some) {
  all.push(...some)
  return all
}

/**
 * Get only unique items in an Array
 */

function unique (results) {
  return Array.from(new Set(results))
}
