const hub = require('./hub')
const getStorage = require('./storage')
const buildResponse = require('./build-response')
const EventEmitter = require('events').EventEmitter
const graphCache = require('./graph-cache')
const async = require('async')
const camelize = require('camelize')
const doNotify = require('./actions/notify')

const Debug = require('debug')

const debugSecurity = Debug('nsolid-console/poll/security')

const Security = require('./security')
const { Application, Package } = require('./models')

const commands = [
  'info',
  'system_info',
  'system_stats',
  'process_stats'
]

module.exports = poll

function poll (config, cache, broadcast) {
  var ee = new EventEmitter()
  ee.on('error', (err) => {
    if (Array.isArray(err)) {
      err.forEach(err => console.error(err.stack))
    } else {
      console.error(err.stack)
    }
  })

  // only on first run
  ee.once('security.updating', () => {
    cache.security.updating = true
  })
  ee.on('security.updated', () => {
    const now = Date.now()

    // Add last updated time to any application without one
    const apps = cache.graph.getIds('applications')
    const updated = apps.map((appid) => {
      const app = cache.graph.data[appid]
      if (!app.lastSecurityUpdate) {
        app.lastSecurityUpdate = now
        return app.name
      }
    }).filter(Boolean)

    debugSecurity('add last updated to applications: %s', updated.join(', '))

    cache.security.updating = false
  })

  // whenever updating all
  ee.on('security.updating.all', () => {
    cache.security.updating = true
  })

  ee.on('security.updated.all', () => {
    const now = Date.now()
    // Add last updated time to all applications
    cache.graph.getIds('applications').forEach((appid) => {
      const app = cache.graph.data[appid]
      app.lastSecurityUpdate = now
    })

    debugSecurity('set last updated on all applications')

    cache.security.updating = false
    cache.security.updated = now
  })

  cache.graph.on('added:vulnerabilities', vulnerabilities => {
    sendVulnerabilityNotifications(cache, config, (err) => {
      if (err) return ee.emit('error', err)
    })
  })

  var pollInterval = config.interval
  var duration = config.duration
  var storagePath = config.storage
  var timer = null
  var stopped = false
  var hubErrors = []

  let graphDB = null

  // kick it off
  if (pollInterval !== 'never') {
    getStorage(storagePath, 'graphs', config, function gotStorage (err, db) {
      if (err) ee.emit('error', err)
      graphDB = db
      graphCache.init(graphDB, duration, run)
    })
  }

  ee.once('stop', onStop => {
    stopped = true
    clearTimeout(timer)
    if (!graphDB) return ee.emit('stopped')
    graphDB.close(err => {
      if (err) ee.emit('error', err)
      ee.emit('stopped')
    })
  })

  if (pollInterval !== false) { // don't run during test
    const debug = Debug('nsolid-console/security')
    initGraph(cache.graph, ee, config, updateSecurity)

    // Security Update Polling Loop
    let securityPoll
    const next = () => {
      if (stopped || pollInterval === false) return
      securityPoll = setTimeout(checkSecurity, 1000)
    }

    const checkSecurity = () => {
      // do nothing if security not enabled or no license
      if (!config.rc.securityConfig.enabled || !config.rc.license.key) return next()

      // do nothing if already updating
      if (cache.security.updating) return next()

      if (config.rc.securityConfig.auto) {
        // auto-update: do nothing if not enough time elapsed
        if ((Date.now() - cache.security.updated) < config.securityinterval) return next()
        debug('auto updating...')
      } else {
        // manual-update: wait for magic signal
        if (cache.security.updated !== -1) return next()
        debug('manual updating...')
      }

      updateSecurity(config)
      .then(() => updateModifiedPackages())
      .then(next)
    }

    checkSecurity()

    ee.on('stopped', () => {
      clearTimeout(securityPoll)
    })
  }

  return ee

  function run () {
    async.mapSeries(commands, performBroadcast, broadcastDone)

    // don't run during test
    if (pollInterval === false) return

    // check whether there are any notifications to send
    sendVulnerabilityNotifications(cache, config, (err) => {
      if (err) return ee.emit('error', err)
    })
  }

  function performBroadcast (command, next) {
    (broadcast || hub.broadcast)(command, config, collectHubErrors)

    function collectHubErrors (errors, r) {
      if (errors && errors.length) {
        errors.forEach((o) => {
          var key = o.hub + o.error.message
          if (hubErrors.indexOf(key) === -1) {
            console.error('POLLING ERROR:', o.error.message)
            hubErrors.push(key)
          }
        })
      }

      next(null, r)
    }
  }

  function broadcastDone (err, commandResults) {
    if (err) {
      return setTimeout(run, pollInterval)
    }

    var results = {}

    commands.forEach(function (name, i) {
      results[name] = clean(commandResults[i])
    })

    ee.emit('input', results)

    cache.collectPollingResult(function collect (pollingCache) {
      const hosts = pollingCache.hosts

      results.info.forEach(function collectInfo (info) {
        const tags = info.reply.tags || []
        const p = results.process_stats.find(p => p.meta.id === info.meta.id)
        if (!p) return
        p.meta.tags = tags
      })

      results.system_info.forEach(function collectSystemInfo (system) {
        const host = system.reply
        const hostname = host.hostname

        if (!hosts[hostname]) {
          hosts[hostname] = host
        }
      })

      results.system_stats.forEach(function collectSystemStats (system) {
        const hostname = system.meta.hostname
        const host = system.reply

        if (!hosts[hostname]) return

        if (!hosts[hostname].stats) {
          hosts[hostname].stats = host
          const totalmem = hosts[hostname].totalmem
          host.memoryPercentageUsed = ((totalmem - host.freemem) / totalmem) * 100
        }
      })

      buildResponse(config, pollingCache, results.process_stats, duration)
    }, pollCycleComplete)

    function pollCycleComplete () {
      ee.emit('output', cache.toJSON())

      // for testing
      if (!stopped && pollInterval !== false) {
        timer = setTimeout(run, pollInterval)
      }
    }
  }

  function updateSecurity (config, packages) {
    if (!config.rc.securityConfig.enabled || !config.rc.license.key) return Promise.resolve() // do nothing if not enabled
    return new Promise(resolve => {
      // ensure events are emitted synchronously
      if (cache.security.updating) return resolve(false) // do nothing if already updating
      if (packages && !packages.length) return resolve(false) // do nothing if no packages to check
      ee.emit('security.updating', packages || [])
      // special event when updating all packages
      if (!packages) ee.emit('security.updating.all')
      resolve(true)
    })
    .then(shouldProceed => {
      if (!shouldProceed) return
      return Security.updateVulnerabilities(config, cache.graph, packages)
    })
    .then(() => {
      ee.emit('security.updated')
      // special event when updating all packages
      if (!packages) ee.emit('security.updated.all')
    })
    .catch(err => ee.emit('error', err))
  }

  function updateModifiedPackages () {
    return loadProcessPackages(config, ee, cache.graph, cache.graph.getNodes('processes'))
  }
}

function clean (a) {
  // Until we can land a fix for https://github.com/nodesource/nsolid-epm-cli/issues/2
  if (!a) {
    return []
  }

  return a.filter(function (o) {
    // TODO: it's likely better to show the errors on the client if something
    //       bad is happening.  In all of the cases I've seen so far it's just
    //       an inability to resolve an ipv6 address (derp)
    return !!o.reply
  }).map(function (o) {
    o.reply = camelize(o.reply)
    return o
  })
}

function initGraph (graph, ee, config, updateSecurity) {
  // skip if in test
  if (config.interval === false) return graph

  const debug = Debug('nsolid-console/data')

  let stopped = false
  ee.once('stopped', () => stopped = true)

  ee.on('input', (results) => {
    if (stopped) return
    const processes = getUniqueProcessesFromResults(results, graph)
    const applications = getApplicationsFromProcesses(processes)
    graph.replace('processes', processes)
    graph.replace('applications', applications)
  })

  ee.on('output', (data) => {
    if (stopped) return
    data.graph = graph.export()
  })

  graph.on('added:applications', applications => {
    debug('Added %s applications', applications)
  })

  graph.on('added:processes', processes => {
    if (stopped) return
    processes.forEach(process => {
      graph.relate(
        'processes', 'applications',
        process.id, Application.id({name: process.app})
      )
    })
  })

  graph.on('added:processes', processes => {
    if (stopped) return
    debug('Added %s processes', processes.length)
    loadProcessPackages(config, ee, graph, processes)
  })

  graph.on('removed:processes', processes => {
    if (stopped) return
    debug('Removed %s processes', processes.length)
  })

  graph.on('removing:process', proc => {
    const modifiedIds = graph.related(proc, 'processes', 'modifiedpackages')
    graph.remove('modifiedpackages', modifiedIds)
  })

  graph.on('removed:applications', applications => {
    if (stopped) return
    debug('Removed %s applications', applications.length)
  })

  graph.on('added:packages', packages => {
    if (stopped) return
    debug('Added %s packages', packages.length)
    updateSecurity(config, packages)
    Security.updateModuleLists(config, graph, packages)
  })

  graph.on('removed:packages', packages => {
    debug('Removed %s packages', packages.length)
  })

  graph.on('added:vulnerabilities', vulnerabilities => {
    debug('Added %s vulnerabilities', vulnerabilities.length)
  })

  graph.on('removed:vulnerabilities', vulnerabilities => {
    debug('Removed %s vulnerabilities', vulnerabilities.length)
  })

  graph.on('added:modulelists', modulelists => {
    debug('Added %s modulelists', modulelists.length)
    loadProcessPackages(config, ee, graph, graph.getNodes('processes'))
  })

  graph.on('removed:modulelists', modulelists => {
    debug('Removed %s modulelists', modulelists.length)
  })

  graph.on('added:modifiedpackages', modifiedPackages => {
    debug('Added %s modifiedPackages', modifiedPackages.length)
    modifiedPackages.forEach(pkg => {
      debug('Modified Package %s:', pkg.packageId)
      debug('Modified Package %s: %s modified files in %s', pkg.packageId, pkg.modules.length, pkg.path)
      pkg.modules.forEach(module => {
        debug('Modified Package %s: %s file: %s', pkg.packageId, module.status, module.path)
      })
    })
  })

  graph.on('added:modifiedpackages', modifiedPackages => {
    debug('Removed %s modifiedPackages', modifiedPackages.length)
  })

  return graph
}

function getUniqueProcessesFromResults (results, graph) {
  const processes = Array.from(Object.keys(results).reduce((processes, key) => {
    results[key].forEach(r => {
      const id = r.meta.id
      const p = processes.get(id) || {}
      if (key === 'process_stats') {
        p.title = r.reply.title
      }
      processes.set(id, Object.assign(p, r.meta))
    })

    return processes
  }, new Map()).values())
  .sort((a, b) => b.pid - a.pid) // ensure order is 'stable'

  // if process has no title, try grab it from the graph no title means process
  // is either coming online or going away though, so perhaps we should filter
  // it out instead. There may be implications from having a different set of
  // processes in tree & the graph.
  processes.forEach(p => {
    const title = p.title && p.title.trim() // ignore whitespace
    if (title) return
    const existingProcess = graph.data[p.id]
    p.title = existingProcess && existingProcess.title || '(no title)' // default title
  })

  return processes
}

function getApplicationsFromProcesses (processes) {
  return Array.from(new Set(processes.map(process => process.app)))
  .sort((a, b) => a.localeCompare(b))
  .map(name => ({name}))
}

function loadProcessPackages (config, ee, graph, processes) {
  if (!config.rc.securityConfig.enabled || !config.rc.license.key) return Promise.resolve() // do nothing if not enabled
  return new Promise(resolve => {
    loadPackages(config, processes, (err, results) => {
      if (err) ee.emit('error', err) // no return. Can be err & results

      // flatten list of packages
      const packages = results
      .map(result => result.reply.packages)
      .filter(Boolean)
      .reduce(flatten, [])

      if (!packages.length) {
        return graph.emit('added:packages', packages)
      }

      graph.add('packages', packages)

      results.forEach((result, index) => {
        // ignore if no packages
        if (!result.reply.packages) return

        const process = result.meta
        const packages = removePackageJSON(result.reply.packages)

        // relate packages with processes
        packages.forEach(pkg => {
          graph.relate(
            'processes', 'packages',
            process.id, Package.id(pkg)
          )
        })

        const modifiedPackages = findModifiedPackages(graph, packages)
        graph.addModifiedPackages(process.id, modifiedPackages)
      })

      resolve()
    })
  })
}

function loadPackages (config, processes, done) {
  hub.commandMany('package_info', processes, config, (err, results) => {
    // ignore err. May be results & err

    if (!results || !results.length) return done(err, [])
    return done(err, results)
  })
}

/**
 * Filter out package.json from packages' modules.
 */

function removePackageJSON (packages) {
  for (let processPkg in packages) {
    if (!packages.modules) continue
    processPkg.modules = processPkg.modules.filter(module => module.path !== 'package.json')
  }

  return packages
}

function findModifiedPackages (graph, processPackages) {
  return processPackages
  // get only packages with modules
  .filter(processPkg => processPkg.modules && processPkg.modules.length)
  // map packages
  .map(processPkg => {
    const packageId = Package.id(processPkg)

    // get canonical package if we have it
    const canonicalPackage = graph.getData(graph.related(packageId, 'packages', 'modulelists'))[0] || {}

    const modifiedModules = findModifiedModules(processPkg.modules, canonicalPackage.modules)

    return {
      packageId: packageId,
      path: processPkg.path,
      modules: modifiedModules
    }
  })
  .filter(p => p.modules.length) // only those with some modified modules
}

/**
 * Find modified modules in `modules` by comparing against `canonicalModules`.
 */

function findModifiedModules (modules, canonicalModules) {
    // we may not know the canonical modules just yet
  if (!canonicalModules || !canonicalModules.length) return []
  return modules
  .map(module => {
    const canonicalModule = canonicalModules.find(m => m.path === module.path)
    if (!canonicalModule) {
      // found an unknown file.
      // would be nice to surface this info but for now, ignore.
      return
    }

    // check if file has been modified
    if (canonicalModule.shasum !== module.shasum) {
      return {
        status: 'modified',
        path: module.path,
        expected: canonicalModule.shasum,
        actual: module.shasum
      }
    }
  })
  .filter(Boolean) // keep only list of modifications
}

/**
 * Array#reduce util.
 * Flattens a singly nested Array.
 */

function flatten (all, some) {
  all.push(...some)
  return all
}

/**
 * Find any `securityConfig.recipients` who have not receieved notifications
 * for the current set of vulnerabilities. Can be called as often as you like,
 * since it won't send notifications unless the email address has never seen a
 * notification for that vulnerability before.
 */

function sendVulnerabilityNotifications (cache, appConfig, done) {
  if (!appConfig.rc.securityConfig.enabled || !appConfig.rc.license.key) return done()
  const debug = Debug('nsolid-console/security')
  const vulnerabilities = cache.graph.getNodes('vulnerabilities')
  if (!vulnerabilities.length) return done() // nothing to do if no vulnerabilities
  const securityConfig = appConfig.rc.securityConfig || {}
  const recipients = securityConfig.recipients = securityConfig.recipients || []

  const sentTo = cache.security.sentNotifications || {}

  // collect payloads for unsent vulnerabilities
  const payloads = recipients.reduce((payloads, recipient) => {
    const sentNotifications = sentTo[recipient] = sentTo[recipient] || []

    // find vulnerabilities to notify about
    const unsentVulns = vulnerabilities.filter(vuln => {
      const notificationId = `vulnerability/${vuln.id}`
      return sentNotifications.indexOf(notificationId) === -1
    })

    // nothing to do if no unsent messages
    if (!unsentVulns.length) return payloads

    const payload = {
      type: 'security',
      method: 'email',
      data: {
        recipients: [recipient],
        vulnerabilities: unsentVulns
      }
    }
    payloads.push(payload)
    return payloads
  }, [])

  payloads.forEach(payload => {
    // Add message id added to list of sent messages for each recipient.
    // Assumes messages are successful until proven otherwise.
    const recipients = payload.data.recipients
    payload.data.vulnerabilities.forEach(vuln => {
      recipients.forEach(recipient => {
        addToSentList(vuln, recipient)
      })
    })
  })

  async.map(payloads, (payload, next) => {
    doNotify(appConfig, payload, (err, data) => {
      if (err) {
        // remove from sent list if there was an error
        const recipients = payload.data.recipients
        recipients.forEach(recipient => {
          payload.data.vulnerabilities.forEach(vuln => {
            removeFromSentList(vuln, recipient)
          })
        })
        return next(err, data)
      }

      debug('notified %s recipients about %s vulnerabilities.', payload.data.recipients, payload.data.vulnerabilities.map(v => v.id))

      next(null, data)
    })
  }, done)

  function addToSentList (vuln, recipient) {
    const notificationId = `vulnerability/${vuln.id}`
    const sentNotifications = sentTo[recipient] || []
    if (sentNotifications.indexOf(notificationId) !== -1) return
    sentNotifications.push(notificationId)
    sentTo[recipient] = sentNotifications
  }

  function removeFromSentList (vuln, recipient) {
    const notificationId = `vulnerability/${vuln.id}`
    const sentNotifications = sentTo[recipient] || []
    const messageIndex = sentNotifications.indexOf(notificationId)
    if (messageIndex === -1) return
    sentNotifications.splice(messageIndex, 1)
    sentTo[recipient] = sentNotifications
  }
}
