const url = require('url')
const path = require('path')
const fs = require('fs')
const createRouter = require('routes')
const getConfig = require('./config')
const concat = require('concat-stream')
const createSnapshot = require('./actions/snapshot')
const createProfile = require('./actions/profile')
const hyperquest = require('hyperquest')

function json (str) {
  try {
    return JSON.parse(str)
  } catch (e) {
    return null
  }
}

module.exports = createServerRoutes

function respondWithError (code, res, err) {
  // TODO: proper error codes
  res.writeHead(code, { 'content-type': 'application/json' })

  const errorObject = {
    message: err.message
  }

  if (err.code) {
    errorObject.code = err.code
  }

  res.end(JSON.stringify(errorObject) + '\n')
}

function respondFileExists (res, file) {
  fs.stat(file, function (e, r) {
    if (e) {
      res.writeHead(404)
    } else {
      res.writeHead(200, {
        'content-length': r.size
      })
    }

    res.end()
  })
}

function createServerRoutes (config, cache, broadcast) {
  const router = createRouter()
  const storagePath = config.storage

  router.addRoute('/heapsnapshot', function createHeapSnapshotRoute (req, res, match) {
    const method = req.method.toLowerCase()
    const parsedUrl = url.parse(req.url, true)
    const params = parsedUrl.query
    const time = params.time || Date.now()
    const filename = cache.generateResourceFilename(
      params.app,
      params.id,
      time,
      'heapsnapshot'
    )

    if (!filename) {
      return respondWithError(400, res, new Error('could not compute filename'))
    }

    const file = path.resolve(storagePath, path.join('snapshots', filename))

    if (method === 'head') {
      respondFileExists(res, file)
    } else if (method === 'get') {
      const readStream = fs.createReadStream(file)
      readStream.on('open', function downloadSnapshotOpen () {
        res.setHeader('Content-Type', 'application/json')
        res.setHeader('Content-Disposition', 'attachment; filename=' + filename)
        readStream.pipe(res)
      })

      readStream.on('error', function downloadSnapshotError (error) {
        respondWithError(404, res, error)
      })
    } else if (method === 'post') {
      const proc = cache.getProcess(params.app, params.id)
      if (!proc) {
        return respondWithError(404, res, new Error('process not found'))
      }

      createSnapshot(cache, config, proc, function snapshotCreated (err, meta) {
        if (err) {
          return respondWithError(500, res, err)
        }

        res.setHeader('content-type', 'application/json')
        res.end(JSON.stringify(meta) + '\n')
      })
    } else {
      res.end()
    }
  })

  router.addRoute('/cpuprofile', function createProfileRoute (req, res, match, db) {
    const method = req.method.toLowerCase()
    const parsedUrl = url.parse(req.url, true)
    const params = parsedUrl.query
    const time = params.time || Date.now()
    const filename = cache.generateResourceFilename(
      params.app,
      params.id,
      time,
      'cpuprofile'
    )

    if (!filename) {
      res.writeHead(400)
      return res.end()
    }

    const file = path.resolve(storagePath, path.join('profiles', filename))

    if (method === 'head') {
      respondFileExists(res, file)
    } else if (method === 'get') {
      // e.g.: http://localhost:3000/applications/default-nsolid-scenario/processes/1062671c412debb5fdd538e08a333bd1f51dd152/profile?time=1438049368666
      const readStream = fs.createReadStream(file)
      readStream.on('open', function () {
        res.setHeader('Content-Type', 'application/json')
        res.setHeader('Content-Disposition', 'attachment; filename=' + filename)

        readStream.pipe(res)
      })

      readStream.on('error', function (error) {
        if (error.code === 'ENOENT') {
          res.writeHead(404)
        } else {
          res.writeHead(500)
        }
        res.end(error.toString())
      })
    } else if (method === 'post') {
      let duration = 5
      if (params.duration) {
        const intermediary = parseInt(params.duration, 10)
        if (!isNaN(intermediary)) {
          duration = intermediary
        }
      }

      const proc = cache.getProcess(params.app, params.id)

      createProfile(cache, config, proc, duration * 1000, function profileCreated (err, meta) {
        if (err) {
          return respondWithError(500, res, err)
        }

        res.setHeader('content-type', 'application/json')
        res.end(JSON.stringify(meta) + '\n')
      })
    } else {
      res.end()
    }
  })

  // delete key
  router.addRoute('/config/license', function (req, res, match) {
    const license = {
      key: '',
      email: '',
      status: 'EMPTY'
    }
    const method = req.method.toLowerCase()
    if (method !== 'delete') {
      respondWithError(501, res, new Error('Method Not Implemented'))
    } else {
      config.rc.license = license

      getConfig.save(config, (e, r) => {
        if (e) {
          console.error(e)
        }
        // notify all of the connected clients
        broadcast({ license })
        res.writeHead(200)
        res.end()
      })
    }
  })

  // add key via put
  router.addRoute('/config/licensekey/:licensekey', function updateLicenseKey (req, res, match) {
    if (req.method.toLowerCase() !== 'put') {
      respondWithError(501, res, new Error('Method Not Implemented'))
    } else { // check the license key
      let statusCode = 0
      hyperquest.get(config.rc.licensekeyEndpoint.url + '/' + match.params.licensekey)
      .on('error', function handleError (error) {
        respondWithError(501, res, error)
      })
      .on('response', function handleNotifyResponse (response) {
        statusCode = response.statusCode
      })
      .pipe(concat((data) => {
        try {
          const body = JSON.parse(data.toString('utf8'))
          if (statusCode === 200) {
            config.rc.license.key = body.key
            config.rc.license.status = body.key.length === 36 ? 'VALID' : 'INVALID'
            getConfig.save(config, (e, r) => {
              if (e) {
                console.error(e)
              }

              // mark security stale so the next polling cycle does a full check
              const securityUpdating = markSecurityStale()

              // notify all of the connected clients
              broadcast(Object.assign({
                license: {
                  key: body.key,
                  status: body.key.length === 36 ? 'VALID' : 'INVALID'
                }
              }, securityUpdating))

              res.writeHead(200)
              res.end()
            })
          } else {
            res.writeHead(statusCode)
            res.end()
          }
        } catch (exception) {
          respondWithError(500, res, new Error('Internal Server Error'))
        }
      }))
    }
  })

  router.addRoute('/config/thresholds/:category/:id', function updateThreshold (req, res, match, db) {
    const method = req.method.toLowerCase()
    switch (method) {
      case 'put':
        req.pipe(concat(buf => {
          const thresholdConfig = json(buf.toString())
          if (!thresholdConfig) {
            return respondWithError(400, res, new Error('Invalid JSON'))
          }
          const rc = config.rc

          rc.thresholds = rc.thresholds || {}

          if (!rc.thresholds[match.params.category]) {
            rc.thresholds[match.params.category] = {}
          }

          const thresholdCategory = rc.thresholds[match.params.category]

          thresholdCategory[match.params.id] = thresholdConfig

          getConfig.save(config, (e, r) => {
            if (e) {
              console.error(e)
            }

            // notify all of the connected clients
            broadcast({
              thresholds: rc.thresholds
            })

            res.writeHead(200)
            res.end()
          })
        }))

        break
      default:
        respondWithError(501, res, new Error('Method Not Implemented'))
        break
    }
  })

  function markSecurityStale () {
    // Reset the last updated time so it will trigger a security
    // update on next poll cycle. A hacky solution for communicating with
    // code that lives in poll.js.
    const oldUpdated = cache.security.updated // keep so can report old value to client
    cache.security.updated = -1

    return {
      security: Object.assign({}, cache.security, {
        updated: oldUpdated,
        updating: true // preemptive state change
      })
    }
  }

  router.addRoute('/security/check-now', function securityCheckNow (req, res, match, db) {
    const method = req.method.toLowerCase()
    if (method !== 'get') return respondWithError(501, res, new Error('Method Not Implemented'))

    // do nothing if already updating
    if (!cache.security.updating) {
      broadcast(markSecurityStale())
    }

    res.writeHead(200)
    res.end()
  })

  router.addRoute('/config/security', function updateSecurity (req, res, match, db) {
    const method = req.method.toLowerCase()
    if (method !== 'put') return respondWithError(501, res, new Error('Method Not Implemented'))
    req.pipe(concat(buf => {
      const data = json(buf.toString())
      if (!data) return respondWithError(400, res, new Error('Invalid JSON'))

      config.rc.securityConfig.recipients = data.recipients
      config.rc.securityConfig.enabled = data.enabled
      config.rc.securityConfig.auto = data.auto

      const isNowEnabled = !cache.security.enabled && data.enabled
      const isNowAuto = !cache.security.auto && data.auto

      // update cache
      Object.assign(cache.security, {
        enabled: data.enabled,
        auto: data.auto,
        recipients: data.recipients
      })

      if (isNowEnabled && isNowAuto && !cache.security.updating) {
        broadcast(markSecurityStale())
      }

      getConfig.save(config, (err) => {
        if (err) console.error(err) // continue on error?

        // notify all connected clients
        broadcast({
          security: cache.security
        })

        res.writeHead(200)
        res.end()
      })
    }))
  })

  return router
}
