'use strict'

import Debug from 'debug'
import * as HTTPClient from 'http-client'
import {Package, Vulnerability} from './models'

const debug = Debug('nsolid-console/security')

const _fetch = HTTPClient.createFetch(
  HTTPClient.accept('application/json'),
  HTTPClient.method('POST'),
  HTTPClient.parseJSON(),
  HTTPClient.requestInfo()
)

const fetch = (input, init = {}) => {
  return _fetch(input, init)
  .then((res) => {
    debug(`${res.url} ${res.status} ${res.statusText}.`)
    return res
  })
  .then(handleResponse)
  .catch((err) => {
    // if error has retryRequest property === false, don't retry
    if (err.retryRequest === false) throw err

    return getRetryOnError(input, init)(err)
  })
}

export function updateVulnerabilities (config, graph, packages) {
  if (!config.rc.securityConfig.enabled || !config.rc.license.key) return Promise.resolve([]) // do nothing if not enabled
  const useAll = !packages
  packages = packages || graph.getNodes('packages')
  if (!packages.length) return Promise.resolve([])

  debug(`Updating vulnerabilities for ${useAll ? 'all ' : ''}%s packages.`, packages.length)

  return fetchSecurityReports(config, packages, { moduleSHAs: false })
  .then(getVulnerabilitiesFromReports)
  .catch(err => {
    // log any errors
    debug(err.message)
    return []
  })
  .then(results => {
    const newVulnerabilities = graph.add('vulnerabilities', results.map(result => result.reply))

    // relate packages with vulnerabilities
    results.forEach(result => {
      graph.relate(
        'packages', 'vulnerabilities',
        Package.id(result.meta), Vulnerability.id(result.reply)
      )
    })

    debug(`Completed: Updating vulnerabilities for ${useAll ? 'all ' : ''}%s packages.`, packages.length)
    return newVulnerabilities
  })
}

export function updateModuleLists (config, graph, packages) {
  if (!config.rc.securityConfig.enabled || !config.rc.license.key) return Promise.resolve([]) // do nothing if not enabled
  const useAll = !packages
  packages = packages || graph.getNodes('packages')
  if (!packages.length) return Promise.resolve([])

  updateModuleLists.inFlight = updateModuleLists.inFlight || new Set()

  const canonicalPackages = new Set(graph.getIds('modulelists').map(id => graph.relatedBack(id, 'modulelists', 'packages')[0]).filter(Boolean))
  const newPackages = packages.filter(pkg => !canonicalPackages.has(pkg.id)) // only packages we haven't already got data for
  const notInflight = newPackages.filter(pkg => !updateModuleLists.inFlight.has(pkg.id)) // only packages we haven't already got data for

  debug(`ModuleLists need updating for %s of ${useAll ? 'all ' : ''}%s packages. Requests for %s packages are currently in flight.`, notInflight.length, packages.length, updateModuleLists.inFlight.size)

  // add to inflight list
  notInflight.forEach(pkg => {
    updateModuleLists.inFlight.add(pkg.id)
  })

  const MAX_PACKAGES = 10 // maximum number of packages to query in one request
  return fetchPackagesChunked(config, notInflight, { moduleSHAs: true }, MAX_PACKAGES)
  .then(getModuleListsFromReports)
  .catch(err => {
    // log any errors
    debug(err.message)
    return []
  })
  .then(moduleLists => {
    // add canonical module paths + shas
    const addedModuleLists = graph.add('modulelists', moduleLists)

    addedModuleLists.forEach(moduleList => {
      graph.relate('packages', 'modulelists', Package.id(moduleList), moduleList.id)
    })

    // remove from inflight list
    notInflight.forEach(pkg => {
      updateModuleLists.inFlight.delete(pkg.id)
    })

    debug(`Completed: Updating ModuleLists for %s of ${useAll ? 'all ' : ''}%s packages. Requests for %s packages are currently in flight.`, notInflight.length, packages.length, updateModuleLists.inFlight.size)
    return moduleLists
  })
}

/**
 * Fetch security reports from server
 */

function fetchSecurityReports (config, packages, options = {}) {
  const securityUrl = config.rc.securityEndpoint.url
  const licenseKey = config.rc.license.key
  debug('Fetching from %s.', securityUrl)
  return fetch(securityUrl, {
    headers: {
      'x-nsolid-license-key': licenseKey
    },
    body: JSON.stringify({
      options,
      packages
    })
  })
  .then(getReports)
}

/**
 * Extract an Array of reports out of a response object.
 */

function handleResponse (res) {
  const message = res.jsonData && res.jsonData.message || ''
  if (res.status === 403) {
    const err = new Error(`${res.url} ${res.status} ${res.statusText}: License key is not valid.`)
    err.retryRequest = false
    throw err
  } else if (res.status === 500) {
    throw new Error(`${res.url} ${res.status} ${res.statusText}: Internal server error getting vulns. ${message}`)
  } else if (res.status === 503) {
    const unavailService = res.jsonData.unavailableService || 'unknown'
    throw new Error(`${res.url} ${res.status} ${res.statusText}: Internal service unavailable: ${unavailService}`)
  } else if (res.status !== 200) {
    throw new Error(`${res.url} ${res.status} ${res.statusText}: Unexpected status code. ${message}`)
  }
  return res
}

/**
 * Extract an Array of reports out of a response object.
 */

function getReports (res) {
  const { reports } = res.jsonData
  if (!reports) {
    const message = res.jsonData && res.jsonData.message || ''
    throw new Error(`Server did not provide any reports. ${message}`)
  }

  return reports
}

/**
 * Extract vulnerabilities from an Array of reports.
 */

function getVulnerabilitiesFromReports (reports) {
  // only deal with packages with vulnerabilities
  const vulnerablePackages = reports.filter(report => report.vulnerabilities.length)
  // flatten vulnerabilities
  return vulnerablePackages.reduce((vulnerabilities, pkg) => {
    const responses = pkg.vulnerabilities.map(vuln => {
      return {
        // keep track of which package was affected
        meta: {
          name: pkg.name,
          version: pkg.version
        },
        reply: vuln
      }
    })
    vulnerabilities.push(...responses)
    return vulnerabilities
  }, [])
}

function getModuleListsFromReports (reports) {
  // only need name, version and modules fields
  return reports.map(({name, version, modules}) => (
    {name, version, modules}
  ))
}

/**
 * Fetch all packages across multiple sumultaneous requests with packages lists of size `maxPackages`.
 */

function fetchPackagesChunked (config, packages, options, maxPackages) {
  return chunkedPromiseAll(packages, maxPackages, subsetOfPackages => {
    return fetchSecurityReports(config, subsetOfPackages, options)
  })
}

/**
 * Apply `mapFn` across `items`, in chunks the size of `chunkSize`.  `mapFn`
 * should return a Promise. All promises created are executed in parallel, but
 * the data sent to each promise is split into smaller chunks.
 */

function chunkedPromiseAll (items, chunkSize, mapFn) {
  return Promise.all(
    chunks(items, chunkSize).map(chunkItems => mapFn(chunkItems))
  )
  .then(results => results.reduce(flatten, []))
}

/**
 * Creates an array of elements split into groups the length of size. If array
 * can’t be split evenly, the final chunk will be the remaining elements.
 */

function chunks (arr, maxSize) {
  if (!maxSize) return []
  return arr.reduce((chunks, pkg, i) => {
    const chunk = chunks[chunks.length - 1] || []
    chunk.push(pkg)
    if (
      !(i % maxSize) &&
      i !== arr.length - 1
    ) chunks.push([])
    return chunks
  }, [])
}

/**
 * Array#reduce util. Flattens a singly nested Array.
 */

function flatten (all, some) {
  all.push(...some)
  return all
}

/**
 * Return .catch function for that will retry the fetch if the request fails.
 */

function getRetryOnError (input, init) {
  init = Object.assign({}, {
    retries: 5, // maximum number of retries
    retryDelay: 1000, // duration to wait before retry
    retryBackoffFactor: 2, // factor to increase delay duration after failure
    retryDelayMax: 30000, // maximum delay after calculating backoff
    randomAmount: 1000 // add some randomness to alieviate congestion
  }, init)
  return function retryOnError (err) {
    --init.retries
    debug(`${input} Failed. Retries left: ${init.retries}`)
    if (init.retries === 0) throw err // run out of tries
    // calculate backoff
    init.retryDelay = Math.round(
      init.randomAmount * Math.random() +
      Math.min(
        init.retryDelayMax,
        init.retryDelay * init.retryBackoffFactor
      )
    )
    debug(`${input} Retrying after %sms`, init.retryDelay)
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(fetch(input, init))
      }, init.retryDelay)
    })
  }
}
