import humanizeDuration from 'humanize-duration'

/**
 * Custom units for 'humanize-duration'.
 */

const shortEnglishHumanizer = humanizeDuration.humanizer({
  language: 'shortEn',
  languages: {
    shortEn: {
      y () { return 'y' },
      mo () { return 'mo' },
      w () { return 'w' },
      d () { return 'd' },
      h () { return 'h' },
      m () { return 'm' },
      s () { return 's' },
      ms () { return 'ms' }
    }
  }
})

/**
 * Format some number of seconds into human-readable format.
 * e.g: "10min 5sec"
 */

function duration (seconds) {
  return shortEnglishHumanizer(seconds * 1000, {
    round: true,
    spacer: '',
    delimiter: ' '
  })
}

/**
 * Two-mode rounding, depending on scale of input:
 * Rounding to two decimal places for numbers between 0 and 1.
 * Normal integer rounding for numbers with absolute value >= 1.
 *
 * e.g.
 * round(0.0123) => 0.01
 * round(123.0123) => 123
 */

function round (num) {
  if (Math.abs(num) >= 1) return Math.round(num)
  else return Math.round(num * 100) / 100
}

/**
 * Standard date output format.
 * @param  {Dateish} date Must be parseable by new Date(date)
 * @return {String}
 */
function date (date) {
  const d = new Date(date)
  const timezone = d.toString().split('(')[1].slice(0, -1)
  return `${d.toLocaleDateString()} ${d.toLocaleTimeString()} ${timezone}`
}

function toMilliseconds (val, units) {
  switch (units) {
    case 'seconds':
      return val * 1000
    case 'minutes':
      return val * 60 * 1000
    case 'hours':
      return val * 60 * 60 * 1000
    default:
      return Number.MAX_VALUE
  }
}

function timerElapsed (start, delay) {
  if (typeof start === 'undefined' || typeof delay === 'undefined') {
    return false
  }

  return start + delay < Date.now()
}

export {
  duration,
  round,
  date,
  toMilliseconds,
  timerElapsed
}
